@@ -1,8 +1,6 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
-const Big = imports.gi.Big;
-const Clutter = imports.gi.Clutter;
-const Pango = imports.gi.Pango;
+const GConf = imports.gi.GConf;
 const GLib = imports.gi.GLib;
 const Gio = imports.gi.Gio;
 const Shell = imports.gi.Shell;
@@ -48,10 +46,14 @@ PlaceInfo.prototype = {
         for (let i = 0; i < terms.length; i++) {
             let term = terms[i];
             let idx = this._lowerName.indexOf(term);
-            if (idx == 0)
-                return Search.MatchType.PREFIX;
-            else if (idx > 0)
-                mtype = Search.MatchType.SUBSTRING;
+            if (idx == 0) {
+                mtype = Search.MatchType.PREFIX;
+            } else if (idx > 0) {
+                if (mtype == Search.MatchType.NONE)
+                    mtype = Search.MatchType.SUBSTRING;
+            } else {
+                return Search.MatchType.NONE;
+            }
         }
         return mtype;
     },
@@ -72,7 +74,7 @@ PlaceDeviceInfo.prototype = {
         this._mount = mount;
         this.name = mount.get_name();
         this._lowerName = this.name.toLowerCase();
-        this.id = "mount:" + mount.get_root().get_uri();
+        this.id = 'mount:' + mount.get_root().get_uri();
     },
 
     iconFactory: function(size) {
@@ -93,12 +95,18 @@ PlaceDeviceInfo.prototype = {
         if (!this.isRemovable())
             return;
 
-        this._mount.unmount(0, null, Lang.bind(this, this._removeFinish));
+        if (this._mount.can_eject())
+            this._mount.eject(0, null, Lang.bind(this, this._removeFinish));
+        else
+            this._mount.unmount(0, null, Lang.bind(this, this._removeFinish));
     },
 
     _removeFinish: function(o, res, data) {
         try {
-            this._mount.unmount_finish(res);
+            if (this._mount.can_eject())
+                this._mount.eject_finish(res);
+            else
+                this._mount.unmount_finish(res);
         } catch (e) {
             let message = _("Failed to unmount '%s'").format(o.get_name());
             Main.overview.infoBar.setMessage(message,
@@ -115,13 +123,13 @@ function PlacesManager() {
 
 PlacesManager.prototype = {
     _init: function() {
-        let gconf = Shell.GConf.get_default();
-        gconf.watch_directory(NAUTILUS_PREFS_DIR);
+        let gconf = GConf.Client.get_default();
+        gconf.add_dir(NAUTILUS_PREFS_DIR, GConf.ClientPreloadType.PRELOAD_NONE);
 
         this._defaultPlaces = [];
         this._mounts = [];
         this._bookmarks = [];
-        this._isDesktopHome = gconf.get_boolean(DESKTOP_IS_HOME_KEY);
+        this._isDesktopHome = gconf.get_bool(DESKTOP_IS_HOME_KEY);
 
         let homeFile = Gio.file_new_for_path (GLib.get_home_dir());
         let homeUri = homeFile.get_uri();
@@ -150,7 +158,7 @@ PlacesManager.prototype = {
 
         this._connect = new PlaceInfo('special:connect', _("Connect to..."),
             function (size) {
-                return St.TextureCache.get_default().load_icon_name("applications-internet", size);
+                return St.TextureCache.get_default().load_icon_name('applications-internet', St.IconType.FULLCOLOR, size);
             },
             function () {
                 new Shell.Process({ args: ['nautilus-connect-server'] }).run();
@@ -163,7 +171,7 @@ PlacesManager.prototype = {
             try {
                 networkApp = Shell.AppSystem.get_default().load_from_desktop_file('network-scheme.desktop');
             } catch(e) {
-                log("Cannot create \"Network\" item, .desktop file not found or corrupt.");
+                log('Cannot create "Network" item, .desktop file not found or corrupt.');
             }
         }
 
@@ -203,7 +211,7 @@ PlacesManager.prototype = {
         this._volumeMonitor.connect('drive-changed', Lang.bind(this, this._updateDevices));
         this._updateDevices();
 
-        this._bookmarksPath = GLib.build_filenamev([GLib.get_home_dir(), ".gtk-bookmarks"]);
+        this._bookmarksPath = GLib.build_filenamev([GLib.get_home_dir(), '.gtk-bookmarks']);
         this._bookmarksFile = Gio.file_new_for_path(this._bookmarksPath);
         let monitor = this._bookmarksFile.monitor_file(Gio.FileMonitorFlags.NONE, null);
         this._bookmarkTimeoutId = 0;
@@ -220,7 +228,7 @@ PlacesManager.prototype = {
 
         this._reloadBookmarks();
 
-        gconf.connect('changed::' + DESKTOP_IS_HOME_KEY, Lang.bind(this, this._updateDesktopMenuVisibility));
+        gconf.notify_add(DESKTOP_IS_HOME_KEY, Lang.bind(this, this._updateDesktopMenuVisibility));
 
     },
 
@@ -331,11 +339,11 @@ PlacesManager.prototype = {
     },
 
     _updateDesktopMenuVisibility: function() {
-        let gconf = Shell.GConf.get_default();
+        let gconf = GConf.Client.get_default();
         this._isDesktopHome = gconf.get_boolean(DESKTOP_IS_HOME_KEY);
 
         if (this._isDesktopHome)
-            this._removeById(this._defaultPlaces, "special:desktop");
+            this._removeById(this._defaultPlaces, 'special:desktop');
         else
             this._defaultPlaces.splice(this._desktopMenuIndex, 0,
                                        this._desktopMenu);
@@ -424,7 +432,7 @@ DashPlaceDisplayItem.prototype = {
         box.add(text, { expand: true, x_fill: true });
 
         if (info.isRemovable()) {
-            let removeIcon = St.TextureCache.get_default().load_icon_name ('media-eject', PLACES_ICON_SIZE);
+            let removeIcon = St.TextureCache.get_default().load_icon_name ('media-eject', St.IconType.FULLCOLOR, PLACES_ICON_SIZE);
             let removeIconBox = new St.Clickable({ child: removeIcon,
                                                    reactive: true });
             box.add(removeIconBox);
@@ -435,18 +443,17 @@ DashPlaceDisplayItem.prototype = {
         }
 
         this.actor.connect('clicked', Lang.bind(this, this._onClicked));
-        this.actor.connect('notify::hover',
-                           Lang.bind(this, this._onHoverChanged));
-        this.actor.connect('button-press-event',
-                           Lang.bind(this, this._onButtonPress));
-        this.actor.connect('button-release-event',
-                           Lang.bind(this, this._onButtonRelease));
 
         this.actor._delegate = this;
-        this._dragStartX = null;
-        this._dragStartY = null;
-        this._draggable = DND.makeDraggable(this.actor,
-                                            { manualMode: true });
+        this._draggable = DND.makeDraggable(this.actor);
+        this._draggable.connect('drag-begin',
+                                Lang.bind(this, function() {
+                                    Main.overview.beginItemDrag(this);
+                                }));
+        this._draggable.connect('drag-end',
+                                Lang.bind(this, function() {
+                                    Main.overview.endItemDrag(this);
+                                }));
     },
 
     _onClicked: function(b) {
@@ -454,36 +461,6 @@ DashPlaceDisplayItem.prototype = {
         Main.overview.hide();
     },
 
-    _onButtonPress: function(actor, event) {
-        if (event.get_button() != 1)
-            return false;
-
-        let [stageX, stageY] = event.get_coords();
-        this._dragStartX = stageX;
-        this._dragStartY = stageY;
-        return false;
-    },
-
-    _onButtonRelease: function(actor, event) {
-        if (event.get_button() != 1)
-            return false;
-
-        this._dragStartX = null;
-        this._dragStartY = null;
-        return false;
-    },
-
-    _onHoverChanged: function(button) {
-        let hover = button.hover;
-        if (!hover) {
-            if (button.pressed && this._dragStartX != null) {
-                button.fake_release();
-                this._draggable.startDrag(this._dragStartX, this._dragStartY,
-                                          global.get_current_time());
-            }
-        }
-    },
-
     getDragActorSource: function() {
         return this._icon;
     },
@@ -510,28 +487,12 @@ DashPlaceDisplay.prototype = {
         // look better in that there would be an even number of items left+right,
         // but it seems like we want some sort of differentiation between actions
         // like "Connect to server..." and regular folders
-        this.actor = new Big.Box({ orientation: Big.BoxOrientation.HORIZONTAL,
-                                   spacing: 4 });
-        this._leftBox = new Big.Box({ orientation: Big.BoxOrientation.VERTICAL });
-        this.actor.append(this._leftBox, Big.BoxPackFlags.EXPAND);
-        this._rightBox = new Big.Box({ orientation: Big.BoxOrientation.VERTICAL });
-        this.actor.append(this._rightBox, Big.BoxPackFlags.EXPAND);
+        this.actor = new St.Table({ style_class: 'places-section',
+                                    homogeneous: true });
 
-        // Subdivide left into actions and devices
-        this._actionsBox = new St.BoxLayout({ style_class: 'places-actions',
-                                               vertical: true });
-
-        this._devBox = new St.BoxLayout({ style_class: 'places-actions',
-                                           name: 'placesDevices',
-                                           vertical: true });
-
-        this._dirsBox = new St.BoxLayout({ style_class: 'places-actions',
-                                            vertical: true });
-
-        this._leftBox.append(this._actionsBox, Big.BoxPackFlags.NONE);
-        this._leftBox.append(this._devBox, Big.BoxPackFlags.NONE);
-
-        this._rightBox.append(this._dirsBox, Big.BoxPackFlags.NONE);
+        this._defaultsList = [];
+        this._bookmarksList = [];
+        this._mountsList = [];
 
         Main.placesManager.connect('defaults-updated', Lang.bind(this, this._updateDefaults));
         Main.placesManager.connect('bookmarks-updated', Lang.bind(this, this._updateBookmarks));
@@ -543,27 +504,40 @@ DashPlaceDisplay.prototype = {
     },
 
     _updateDefaults: function() {
-        this._actionsBox.destroy_children();
+        for (let i = 0; i < this._defaultsList.length; i++)
+            this._defaultsList[i].destroy();
 
+        this._defaultsList = [];
         let places = Main.placesManager.getDefaultPlaces();
-        for (let i = 0; i < places.length; i++)
-            this._actionsBox.add(new DashPlaceDisplayItem(places[i]).actor);
+        for (let i = 0; i < places.length; i++) {
+            this._defaultsList[i] = new DashPlaceDisplayItem(places[i]).actor;
+            this.actor.add(this._defaultsList[i], {row: i, col: 0});
+        }
+        this._updateMounts();
     },
 
     _updateMounts: function() {
-        this._devBox.destroy_children();
+        for (let i = 0; i < this._mountsList.length; i++)
+            this._mountsList[i].destroy();
 
+        this._mountsList = [];
         let places = Main.placesManager.getMounts();
-        for (let i = 0; i < places.length; i++)
-            this._devBox.add(new DashPlaceDisplayItem(places[i]).actor);
+        for (let i = 0; i < places.length; i++) {
+             this._mountsList[i] = new DashPlaceDisplayItem(places[i]).actor;
+             this.actor.add(this._mountsList[i], {row: this._defaultsList.length + i, col: 0});
+        }
     },
 
     _updateBookmarks: function() {
-        this._dirsBox.destroy_children();
+        for (let i = 0; i < this._bookmarksList.length; i++)
+            this._bookmarksList[i].destroy();
 
+        this._bookmarksList = [];
         let places = Main.placesManager.getBookmarks();
-        for (let i = 0; i < places.length; i ++)
-            this._dirsBox.add(new DashPlaceDisplayItem(places[i]).actor);
+        for (let i = 0; i < places.length; i ++) {
+            this._bookmarksList[i] = new DashPlaceDisplayItem(places[i]).actor;
+            this.actor.add(this._bookmarksList[i], {row: i, col: 1});
+        }
     }
 };
 
@@ -601,8 +575,9 @@ PlaceSearchProvider.prototype = {
     },
 
     _searchPlaces: function(places, terms) {
-        let multipleResults = [];
+        let multiplePrefixResults = [];
         let prefixResults = [];
+        let multipleSubstringResults = [];
         let substringResults = [];
 
         terms = terms.map(String.toLowerCase);
@@ -610,17 +585,20 @@ PlaceSearchProvider.prototype = {
         for (let i = 0; i < places.length; i++) {
             let place = places[i];
             let mtype = place.matchTerms(terms);
-            if (mtype == Search.MatchType.MULTIPLE)
-                multipleResults.push(place.id);
+            if (mtype == Search.MatchType.MULTIPLE_PREFIX)
+                multiplePrefixResults.push(place.id);
             else if (mtype == Search.MatchType.PREFIX)
                 prefixResults.push(place.id);
+            else if (mtype == Search.MatchType.MULTIPLE_SUBSTRING)
+                multipleSubstringResults.push(place.id);
             else if (mtype == Search.MatchType.SUBSTRING)
                 substringResults.push(place.id);
         }
-        multipleResults.sort(this._compareResultMeta);
+        multiplePrefixResults.sort(this._compareResultMeta);
         prefixResults.sort(this._compareResultMeta);
+        multipleSubstringResults.sort(this._compareResultMeta);
         substringResults.sort(this._compareResultMeta);
-        return multipleResults.concat(prefixResults.concat(substringResults));
+        return multiplePrefixResults.concat(prefixResults.concat(multipleSubstringResults.concat(substringResults)));
     },
 
     getInitialResultSet: function(terms) {
