@@ -25,12 +25,11 @@ static void shell_generic_container_iface_init (ClutterContainerIface *iface);
 
 G_DEFINE_TYPE_WITH_CODE(ShellGenericContainer,
                         shell_generic_container,
-                        ST_TYPE_WIDGET,
+                        ST_TYPE_CONTAINER,
                         G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_CONTAINER,
                                                shell_generic_container_iface_init));
 
 struct _ShellGenericContainerPrivate {
-  GList *children;
   GHashTable *skip_paint;
 };
 
@@ -95,6 +94,8 @@ shell_generic_container_get_preferred_width (ClutterActor *actor,
   if (natural_width_p)
     *natural_width_p = alloc->natural_size;
   shell_generic_container_allocation_unref (alloc);
+
+  st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
 }
 
 static void
@@ -116,17 +117,20 @@ shell_generic_container_get_preferred_height (ClutterActor *actor,
   if (natural_height_p)
     *natural_height_p = alloc->natural_size;
   shell_generic_container_allocation_unref (alloc);
+
+  st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
 }
 
 static void
 shell_generic_container_paint (ClutterActor  *actor)
 {
   ShellGenericContainer *self = (ShellGenericContainer*) actor;
-  GList *iter;
+  GList *iter, *children;
 
   CLUTTER_ACTOR_CLASS (shell_generic_container_parent_class)->paint (actor);
 
-  for (iter = self->priv->children; iter; iter = iter->next)
+  children = st_container_get_children_list (ST_CONTAINER (actor));
+  for (iter = children; iter; iter = iter->next)
     {
       ClutterActor *child = iter->data;
 
@@ -142,11 +146,12 @@ shell_generic_container_pick (ClutterActor        *actor,
                               const ClutterColor  *color)
 {
   ShellGenericContainer *self = (ShellGenericContainer*) actor;
-  GList *iter;
+  GList *iter, *children;
 
   CLUTTER_ACTOR_CLASS (shell_generic_container_parent_class)->pick (actor, color);
 
-  for (iter = self->priv->children; iter; iter = iter->next)
+  children = st_container_get_children_list (ST_CONTAINER (actor));
+  for (iter = children; iter; iter = iter->next)
     {
       ClutterActor *child = iter->data;
 
@@ -157,9 +162,28 @@ shell_generic_container_pick (ClutterActor        *actor,
     }
 }
 
+static GList *
+shell_generic_container_get_focus_chain (StContainer *container)
+{
+  ShellGenericContainer *self = SHELL_GENERIC_CONTAINER (container);
+  GList *children, *focus_chain;
+
+  focus_chain = NULL;
+  for (children = st_container_get_children_list (container); children; children = children->next)
+    {
+      ClutterActor *child = children->data;
+
+      if (CLUTTER_ACTOR_IS_VISIBLE (child) &&
+          !shell_generic_container_get_skip_paint (self, child))
+        focus_chain = g_list_prepend (focus_chain, child);
+    }
+
+  return g_list_reverse (focus_chain);
+}
+
 /**
  * shell_generic_container_get_n_skip_paint:
- * @container:  A #ShellGenericContainer
+ * @self:  A #ShellGenericContainer
  *
  * Returns: Number of children which will not be painted.
  */
@@ -170,10 +194,26 @@ shell_generic_container_get_n_skip_paint (ShellGenericContainer  *self)
 }
 
 /**
- * shell_generic_container_set_skip_paint:
- * @container: A #ShellGenericContainer
+ * shell_generic_container_get_skip_paint:
+ * @self: A #ShellGenericContainer
  * @child: Child #ClutterActor
- * @skip %TRUE if we should skip painting
+ *
+ * Gets whether or not @actor is skipped when painting.
+ *
+ * Return value: %TRUE or %FALSE
+ */
+gboolean
+shell_generic_container_get_skip_paint (ShellGenericContainer  *self,
+                                        ClutterActor           *child)
+{
+  return g_hash_table_lookup (self->priv->skip_paint, child) != NULL;
+}
+
+/**
+ * shell_generic_container_set_skip_paint:
+ * @self: A #ShellGenericContainer
+ * @child: Child #ClutterActor
+ * @skip: %TRUE if we should skip painting
  *
  * Set whether or not we should skip painting @actor.  Workaround for
  * lack of gjs ability to override _paint vfunc.
@@ -195,29 +235,6 @@ shell_generic_container_set_skip_paint (ShellGenericContainer  *self,
     g_hash_table_insert (self->priv->skip_paint, child, child);
 }
 
-/**
- * shell_generic_container_remove_all:
- * @self: A #ShellGenericContainer
- *
- * Removes all child actors from @self.
- */
-void
-shell_generic_container_remove_all (ShellGenericContainer *self)
-{
-  /* copied from clutter_group_remove_all() */
-
-  GList *children;
-
-  children = self->priv->children;
-  while (children)
-    {
-      ClutterActor *child = children->data;
-      children = children->next;
-
-      clutter_container_remove_actor (CLUTTER_CONTAINER (self), child);
-    }
-}
-
 static void
 shell_generic_container_finalize (GObject *object)
 {
@@ -228,25 +245,14 @@ shell_generic_container_finalize (GObject *object)
   G_OBJECT_CLASS (shell_generic_container_parent_class)->finalize (object);
 }
 
-static void
-shell_generic_container_dispose (GObject *object)
-{
-  ShellGenericContainerPrivate *priv = SHELL_GENERIC_CONTAINER (object)->priv;
-
-  while (priv->children)
-    clutter_actor_destroy (priv->children->data);
-
-  G_OBJECT_CLASS (shell_generic_container_parent_class)->dispose (object);
-}
-
 static void
 shell_generic_container_class_init (ShellGenericContainerClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
+  StContainerClass *container_class = ST_CONTAINER_CLASS (klass);
 
   gobject_class->finalize = shell_generic_container_finalize;
-  gobject_class->dispose = shell_generic_container_dispose;
 
   actor_class->get_preferred_width = shell_generic_container_get_preferred_width;
   actor_class->get_preferred_height = shell_generic_container_get_preferred_height;
@@ -254,6 +260,8 @@ shell_generic_container_class_init (ShellGenericContainerClass *klass)
   actor_class->paint = shell_generic_container_paint;
   actor_class->pick = shell_generic_container_pick;
 
+  container_class->get_focus_chain = shell_generic_container_get_focus_chain;
+
   shell_generic_container_signals[GET_PREFERRED_WIDTH] =
     g_signal_new ("get-preferred-width",
                   G_TYPE_FROM_CLASS (klass),
@@ -285,73 +293,18 @@ shell_generic_container_class_init (ShellGenericContainerClass *klass)
 }
 
 static void
-shell_generic_container_add_actor (ClutterContainer *container,
-                                   ClutterActor     *actor)
-{
-  ShellGenericContainerPrivate *priv = SHELL_GENERIC_CONTAINER (container)->priv;
-
-  _st_container_add_actor (container, actor, &priv->children);
-}
-
-static void
-shell_generic_container_remove_actor (ClutterContainer *container,
-                                      ClutterActor     *actor)
+shell_generic_container_actor_removed (ClutterContainer *container,
+                                       ClutterActor     *actor)
 {
   ShellGenericContainerPrivate *priv = SHELL_GENERIC_CONTAINER (container)->priv;
 
   g_hash_table_remove (priv->skip_paint, actor);
-
-  _st_container_remove_actor (container, actor, &priv->children);
-}
-
-static void
-shell_generic_container_foreach (ClutterContainer *container,
-                                 ClutterCallback   callback,
-                                 gpointer          callback_data)
-{
-  ShellGenericContainerPrivate *priv = SHELL_GENERIC_CONTAINER (container)->priv;
-
-  _st_container_foreach (container, callback, callback_data,
-                         &priv->children);
-}
-
-static void
-shell_generic_container_lower (ClutterContainer *container,
-                               ClutterActor     *actor,
-                               ClutterActor     *sibling)
-{
-  ShellGenericContainerPrivate *priv = SHELL_GENERIC_CONTAINER (container)->priv;
-
-  _st_container_lower (container, actor, sibling, &priv->children);
-}
-
-static void
-shell_generic_container_raise (ClutterContainer *container,
-                        ClutterActor     *actor,
-                        ClutterActor     *sibling)
-{
-  ShellGenericContainerPrivate *priv = SHELL_GENERIC_CONTAINER (container)->priv;
-
-  _st_container_raise (container, actor, sibling, &priv->children);
-}
-
-static void
-shell_generic_container_sort_depth_order (ClutterContainer *container)
-{
-  ShellGenericContainerPrivate *priv = SHELL_GENERIC_CONTAINER (container)->priv;
-
-  _st_container_sort_depth_order (container, &priv->children);
 }
 
 static void
 shell_generic_container_iface_init (ClutterContainerIface *iface)
 {
-  iface->add = shell_generic_container_add_actor;
-  iface->remove = shell_generic_container_remove_actor;
-  iface->foreach = shell_generic_container_foreach;
-  iface->lower = shell_generic_container_lower;
-  iface->raise = shell_generic_container_raise;
-  iface->sort_depth_order = shell_generic_container_sort_depth_order;
+  iface->actor_removed = shell_generic_container_actor_removed;
 }
 
 static void
