@@ -29,192 +29,15 @@
 */
 
 #include <stdlib.h>
-#include <string.h>
 #include <math.h>
 
 #include "st-shadow.h"
+#include "st-private.h"
 #include "st-theme-private.h"
 #include "st-theme-context.h"
 #include "st-texture-cache.h"
 #include "st-theme-node-private.h"
 
-/*****
- * Shadows
- *****/
-
-static gdouble *
-calculate_gaussian_kernel (gdouble   sigma,
-                           guint     n_values)
-{
-  gdouble *ret, sum;
-  gdouble exp_divisor;
-  gint half, i;
-
-  g_return_val_if_fail ((int) sigma > 0, NULL);
-
-  half = n_values / 2;
-
-  ret = g_malloc (n_values * sizeof (gdouble));
-  sum = 0.0;
-
-  exp_divisor = 2 * sigma * sigma;
-
-  /* n_values of 1D Gauss function */
-  for (i = 0; i < n_values; i++)
-    {
-      ret[i] = exp (-(i - half) * (i - half) / exp_divisor);
-      sum += ret[i];
-    }
-
-  /* normalize */
-  for (i = 0; i < n_values; i++)
-    ret[i] /= sum;
-
-  return ret;
-}
-
-static CoglHandle
-create_shadow_material (StThemeNode  *node,
-                        CoglHandle    src_texture)
-{
-  CoglHandle  material;
-  CoglHandle  texture;
-  StShadow   *shadow_spec;
-  guchar     *pixels_in, *pixels_out;
-  gint        width_in, height_in, rowstride_in;
-  gint        width_out, height_out, rowstride_out;
-  float       sigma;
-
-  shadow_spec = st_theme_node_get_shadow (node);
-  if (!shadow_spec)
-    return COGL_INVALID_HANDLE;
-
-  /* we use an approximation of the sigma - blur radius relationship used
-     in Firefox for doing SVG blurs; see
-     http://mxr.mozilla.org/mozilla-central/source/gfx/thebes/src/gfxBlur.cpp#280
-  */
-  sigma = shadow_spec->blur / 1.9;
-
-  width_in  = cogl_texture_get_width  (src_texture);
-  height_in = cogl_texture_get_height (src_texture);
-  rowstride_in = (width_in + 3) & ~3;
-
-  pixels_in  = g_malloc0 (rowstride_in * height_in);
-
-  cogl_texture_get_data (src_texture, COGL_PIXEL_FORMAT_A_8,
-                         rowstride_in, pixels_in);
-
-  if ((guint) shadow_spec->blur == 0)
-    {
-      width_out  = width_in;
-      height_out = height_in;
-      rowstride_out = rowstride_in;
-      pixels_out = g_memdup (pixels_in, rowstride_out * height_out);
-    }
-  else
-    {
-      gdouble *kernel;
-      guchar  *line;
-      gint     n_values, half;
-      gint     x_in, y_in, x_out, y_out, i;
-
-      n_values = (gint) 5 * sigma;
-      half = n_values / 2;
-
-      width_out  = width_in  + 2 * half;
-      height_out = height_in + 2 * half;
-      rowstride_out = (width_out + 3) & ~3;
-
-      pixels_out = g_malloc0 (rowstride_out * height_out);
-      line       = g_malloc0 (rowstride_out);
-
-      kernel = calculate_gaussian_kernel (sigma, n_values);
-
-      /* vertical blur */
-      for (x_in = 0; x_in < width_in; x_in++)
-        for (y_out = 0; y_out < height_out; y_out++)
-          {
-            guchar *pixel_in, *pixel_out;
-            gint i0, i1;
-
-            y_in = y_out - half;
-
-            /* We read from the source at 'y = y_in + i - half'; clamp the
-             * full i range [0, n_values) so that y is in [0, height_in).
-             */
-            i0 = MAX (half - y_in, 0);
-            i1 = MIN (height_in + half - y_in, n_values);
-
-            pixel_in  =  pixels_in + (y_in + i0 - half) * rowstride_in + x_in;
-            pixel_out =  pixels_out + y_out * rowstride_out + (x_in + half);
-
-            for (i = i0; i < i1; i++)
-              {
-                *pixel_out += *pixel_in * kernel[i];
-                pixel_in += rowstride_in;
-              }
-          }
-
-      /* horizontal blur */
-      for (y_out = 0; y_out < height_out; y_out++)
-        {
-          memcpy (line, pixels_out + y_out * rowstride_out, rowstride_out);
-
-          for (x_out = 0; x_out < width_out; x_out++)
-            {
-              gint i0, i1;
-              guchar *pixel_out, *pixel_in;
-
-              /* We read from the source at 'x = x_out + i - half'; clamp the
-               * full i range [0, n_values) so that x is in [0, width_out).
-               */
-              i0 = MAX (half - x_out, 0);
-              i1 = MIN (width_out + half - x_out, n_values);
-
-              pixel_in  = line + x_out + i0 - half;
-              pixel_out = pixels_out + rowstride_out * y_out + x_out;
-
-              *pixel_out = 0;
-              for (i = i0; i < i1; i++)
-                {
-                  *pixel_out += *pixel_in * kernel[i];
-                  pixel_in++;
-                }
-            }
-        }
-      g_free (kernel);
-      g_free (line);
-    }
-
-  texture = cogl_texture_new_from_data (width_out,
-                                        height_out,
-                                        COGL_TEXTURE_NONE,
-                                        COGL_PIXEL_FORMAT_A_8,
-                                        COGL_PIXEL_FORMAT_A_8,
-                                        rowstride_out,
-                                        pixels_out);
-
-  g_free (pixels_in);
-  g_free (pixels_out);
-
-  material = cogl_material_new ();
-
-  cogl_material_set_layer (material, 0, texture);
-
-  /* We set up the material to blend the shadow texture with the combine
-   * constant, but defer setting the latter until painting, so that we can
-   * take the actor's overall opacity into account. */
-  cogl_material_set_layer_combine (material, 0,
-                                   "RGBA = MODULATE (CONSTANT, TEXTURE[A])",
-                                   NULL);
-
-
-  cogl_handle_unref (texture);
-
-  return material;
-}
-
-
 /****
  * Rounded corners
  ****/
@@ -229,7 +52,7 @@ typedef struct {
 } StCornerSpec;
 
 static CoglHandle
-create_corner_texture (StCornerSpec *corner)
+create_corner_material (StCornerSpec *corner)
 {
   CoglHandle texture;
   cairo_t *cr;
@@ -344,7 +167,7 @@ load_corner (StTextureCache  *cache,
 {
   LoadCornerData *data = datap;
 
-  return create_corner_texture (data->corner);
+  return create_corner_material (data->corner);
 }
 
 /* To match the CSS specification, we want the border to look like it was
@@ -399,7 +222,7 @@ static CoglHandle
 st_theme_node_lookup_corner (StThemeNode    *node,
                              StCorner        corner_id)
 {
-  CoglHandle texture;
+  CoglHandle texture, material;
   char *key;
   StTextureCache *cache;
   StCornerSpec corner;
@@ -446,10 +269,12 @@ st_theme_node_lookup_corner (StThemeNode    *node,
   data.node = node;
   data.corner = &corner;
   texture = st_texture_cache_load (cache, key, ST_TEXTURE_CACHE_POLICY_NONE, load_corner, &data, NULL);
+  material = _st_create_texture_material (texture);
+  cogl_handle_unref (texture);
 
   g_free (key);
 
-  return texture;
+  return material;
 }
 
 static void
@@ -466,8 +291,8 @@ get_background_position (StThemeNode             *self,
   w = cogl_texture_get_width (self->background_texture);
   h = cogl_texture_get_height (self->background_texture);
 
-  /* scale the background into the allocated bounds */
-  if (w > result->x2 || h > result->y2)
+  /* scale the background into the allocated bounds, when not being absolutely positioned */
+  if ((w > result->x2 || h > result->y2) && !self->background_position_set)
     {
       gint new_h, new_w, offset;
       gint box_w, box_h;
@@ -500,9 +325,18 @@ get_background_position (StThemeNode             *self,
     }
   else
     {
-      /* center the background on the widget */
-      result->x1 = (int)(((allocation->x2 - allocation->x1) / 2) - (w / 2));
-      result->y1 = (int)(((allocation->y2 - allocation->y1) / 2) - (h / 2));
+      /* honor the specified position if any */
+      if (self->background_position_set)
+        {
+          result->x1 = self->background_position_x;
+          result->y1 = self->background_position_y;
+        }
+      else
+        {
+          /* center the background on the widget */
+          result->x1 = (int)(((allocation->x2 - allocation->x1) / 2) - (w / 2));
+          result->y1 = (int)(((allocation->y2 - allocation->y1) / 2) - (h / 2));
+        }
       result->x2 = result->x1 + w;
       result->y2 = result->y1 + h;
     }
@@ -671,25 +505,42 @@ st_theme_node_render_gradient (StThemeNode *node)
 void
 _st_theme_node_free_drawing_state (StThemeNode  *node)
 {
+  int corner_id;
+
   if (node->background_texture != COGL_INVALID_HANDLE)
     cogl_handle_unref (node->background_texture);
+  if (node->background_material != COGL_INVALID_HANDLE)
+    cogl_handle_unref (node->background_material);
   if (node->background_shadow_material != COGL_INVALID_HANDLE)
     cogl_handle_unref (node->background_shadow_material);
   if (node->border_texture != COGL_INVALID_HANDLE)
     cogl_handle_unref (node->border_texture);
+  if (node->border_material != COGL_INVALID_HANDLE)
+    cogl_handle_unref (node->border_material);
   if (node->border_shadow_material != COGL_INVALID_HANDLE)
     cogl_handle_unref (node->border_shadow_material);
 
+  for (corner_id = 0; corner_id < 4; corner_id++)
+    if (node->corner_material[corner_id] != COGL_INVALID_HANDLE)
+      cogl_handle_unref (node->corner_material[corner_id]);
+
   _st_theme_node_init_drawing_state (node);
 }
 
 void
 _st_theme_node_init_drawing_state (StThemeNode *node)
 {
+  int corner_id;
+
   node->background_texture = COGL_INVALID_HANDLE;
+  node->background_material = COGL_INVALID_HANDLE;
   node->background_shadow_material = COGL_INVALID_HANDLE;
   node->border_shadow_material = COGL_INVALID_HANDLE;
   node->border_texture = COGL_INVALID_HANDLE;
+  node->border_material = COGL_INVALID_HANDLE;
+
+  for (corner_id = 0; corner_id < 4; corner_id++)
+    node->corner_material[corner_id] = COGL_INVALID_HANDLE;
 }
 
 static void st_theme_node_paint_borders (StThemeNode           *node,
@@ -738,10 +589,16 @@ st_theme_node_render_resources (StThemeNode   *node,
       node->border_texture = st_theme_node_render_gradient (node);
     }
 
+  if (node->border_texture)
+    node->border_material = _st_create_texture_material (node->border_texture);
+  else
+    node->border_material = COGL_INVALID_HANDLE;
+
   if (shadow_spec)
     {
       if (node->border_texture != COGL_INVALID_HANDLE)
-        node->border_shadow_material = create_shadow_material (node, node->border_texture);
+        node->border_shadow_material = _st_create_shadow_material (shadow_spec,
+                                                                   node->border_texture);
       else if (node->background_color.alpha > 0 ||
                node->border_width[ST_SIDE_TOP] > 0 ||
                node->border_width[ST_SIDE_LEFT] > 0 ||
@@ -766,8 +623,8 @@ st_theme_node_render_resources (StThemeNode   *node,
               cogl_pop_framebuffer ();
               cogl_handle_unref (offscreen);
 
-              node->border_shadow_material = create_shadow_material (node,
-                                                                     buffer);
+              node->border_shadow_material = _st_create_shadow_material (shadow_spec,
+                                                                         buffer);
             }
           cogl_handle_unref (buffer);
         }
@@ -778,66 +635,35 @@ st_theme_node_render_resources (StThemeNode   *node,
     {
 
       node->background_texture = st_texture_cache_load_file_to_cogl_texture (texture_cache, background_image);
+      node->background_material = _st_create_texture_material (node->background_texture);
 
       if (shadow_spec)
         {
-          node->background_shadow_material = create_shadow_material (node, node->background_texture);
+          node->background_shadow_material = _st_create_shadow_material (shadow_spec,
+                                                                         node->background_texture);
         }
     }
 
-  node->corner_texture = st_theme_node_lookup_corner (node, ST_CORNER_TOPLEFT);
+  node->corner_material[ST_CORNER_TOPLEFT] =
+    st_theme_node_lookup_corner (node, ST_CORNER_TOPLEFT);
+  node->corner_material[ST_CORNER_TOPRIGHT] =
+    st_theme_node_lookup_corner (node, ST_CORNER_TOPRIGHT);
+  node->corner_material[ST_CORNER_BOTTOMRIGHT] =
+    st_theme_node_lookup_corner (node, ST_CORNER_BOTTOMRIGHT);
+  node->corner_material[ST_CORNER_BOTTOMLEFT] =
+    st_theme_node_lookup_corner (node, ST_CORNER_BOTTOMLEFT);
 }
 
 static void
-paint_texture_with_opacity (CoglHandle       texture,
-                            ClutterActorBox *box,
-                            guint8           paint_opacity)
+paint_material_with_opacity (CoglHandle       material,
+                             ClutterActorBox *box,
+                             guint8           paint_opacity)
 {
-  if (paint_opacity == 255)
-    {
-      /* Minor: optimization use the default material if we can */
-      cogl_set_source_texture (texture);
-      cogl_rectangle (box->x1, box->y1, box->x2, box->y2);
-      return;
-    }
-
-  CoglHandle material;
-
-  material = cogl_material_new ();
-  cogl_material_set_layer (material, 0, texture);
   cogl_material_set_color4ub (material,
                               paint_opacity, paint_opacity, paint_opacity, paint_opacity);
 
   cogl_set_source (material);
   cogl_rectangle (box->x1, box->y1, box->x2, box->y2);
-
-  cogl_handle_unref (material);
-}
-
-static void
-paint_shadow_with_opacity (CoglHandle       shadow_material,
-                           StShadow        *shadow_spec,
-                           ClutterActorBox *box,
-                           guint8           paint_opacity)
-{
-  ClutterActorBox shadow_box;
-  CoglColor       color;
-
-  st_shadow_get_box (shadow_spec, box, &shadow_box);
-
-  cogl_color_set_from_4ub (&color,
-                           shadow_spec->color.red   * paint_opacity / 255,
-                           shadow_spec->color.green * paint_opacity / 255,
-                           shadow_spec->color.blue  * paint_opacity / 255,
-                           shadow_spec->color.alpha * paint_opacity / 255);
-  cogl_color_premultiply (&color);
-
-  cogl_material_set_layer_combine_constant (shadow_material, 0, &color);
-
-  cogl_set_source (shadow_material);
-  cogl_rectangle_with_texture_coords (shadow_box.x1, shadow_box.y1,
-                                      shadow_box.x2, shadow_box.y2,
-                                      0, 0, 1, 1);
 }
 
 static void
@@ -848,23 +674,29 @@ st_theme_node_paint_borders (StThemeNode           *node,
 {
   float width, height;
   int border_width;
-  int border_radius;
-  int max_width_radius;
+  int max_border_radius = 0;
+  int max_width_radius[4];
+  int corner_id;
   ClutterColor border_color;
-  CoglHandle material;
 
   width = box->x2 - box->x1;
   height = box->y2 - box->y1;
 
   get_arbitrary_border (node, &border_width, &border_color);
-  border_radius = node->border_radius[ST_CORNER_TOPLEFT];
-
-  max_width_radius = MAX(border_width, border_radius);
+  for (corner_id = 0; corner_id < 4; corner_id++)
+    {
+      if (node->border_radius[corner_id] > max_border_radius)
+        max_border_radius = node->border_radius[corner_id];
+      max_width_radius[corner_id] = MAX(border_width,
+                                        node->border_radius[corner_id]);
+    }
 
   /* borders */
   if (border_width > 0)
     {
       ClutterColor effective_border;
+      gboolean skip_corner_1, skip_corner_2;
+      float x1, y1, x2, y2;
 
       over (&border_color, &node->background_color, &effective_border);
 
@@ -873,60 +705,87 @@ st_theme_node_paint_borders (StThemeNode           *node,
                                 effective_border.blue,
                                 paint_opacity * effective_border.alpha / 255);
 
-      if (border_radius > 0) /* skip corners */
-        {
-          /* NORTH */
-          cogl_rectangle (max_width_radius, 0,
-                          width - max_width_radius, border_width);
+      /* NORTH */
+      skip_corner_1 = node->border_radius[ST_CORNER_TOPLEFT] > 0;
+      skip_corner_2 = node->border_radius[ST_CORNER_TOPRIGHT] > 0;
 
-          /* EAST */
-          cogl_rectangle (width - border_width, max_width_radius,
-                          width, height - max_width_radius);
+      x1 = skip_corner_1 ? max_width_radius[ST_CORNER_TOPLEFT] : 0;
+      y1 = 0;
+      x2 = skip_corner_2 ? width - max_width_radius[ST_CORNER_TOPRIGHT] : width;
+      y2 = border_width;
+      cogl_rectangle (x1, y1, x2, y2);
 
-          /* SOUTH */
-          cogl_rectangle (max_width_radius, height - border_width,
-                          width - max_width_radius, height);
+      /* EAST */
+      skip_corner_1 = node->border_radius[ST_CORNER_TOPRIGHT] > 0;
+      skip_corner_2 = node->border_radius[ST_CORNER_BOTTOMRIGHT] > 0;
 
-          /* WEST */
-          cogl_rectangle (0, max_width_radius,
-                          border_width, height - max_width_radius);
-        }
-      else /* include corners */
-        {
-          /* NORTH */
-          cogl_rectangle (0, 0,
-                          width, border_width);
+      x1 = width - border_width;
+      y1 = skip_corner_1 ? max_width_radius[ST_CORNER_TOPRIGHT] : border_width;
+      x2 = width;
+      y2 = skip_corner_2 ? height - max_width_radius[ST_CORNER_BOTTOMRIGHT]
+                         : height - border_width;
+      cogl_rectangle (x1, y1, x2, y2);
 
-          /* EAST */
-          cogl_rectangle (width - border_width, border_width,
-                          width, height - border_width);
+      /* SOUTH */
+      skip_corner_1 = node->border_radius[ST_CORNER_BOTTOMLEFT] > 0;
+      skip_corner_2 = node->border_radius[ST_CORNER_BOTTOMRIGHT] > 0;
 
-          /* SOUTH */
-          cogl_rectangle (0, height - border_width,
-                          width, height);
+      x1 = skip_corner_1 ? max_width_radius[ST_CORNER_BOTTOMLEFT] : 0;
+      y1 = height - border_width;
+      x2 = skip_corner_2 ? width - max_width_radius[ST_CORNER_BOTTOMRIGHT]
+                         : width;
+      y2 = height;
+      cogl_rectangle (x1, y1, x2, y2);
 
-          /* WEST */
-          cogl_rectangle (0, border_width,
-                          border_width, height - border_width);
-        }
+      /* WEST */
+      skip_corner_1 = node->border_radius[ST_CORNER_TOPLEFT] > 0;
+      skip_corner_2 = node->border_radius[ST_CORNER_BOTTOMLEFT] > 0;
+
+      x1 = 0;
+      y1 = skip_corner_1 ? max_width_radius[ST_CORNER_TOPLEFT] : border_width;
+      x2 = border_width;
+      y2 = skip_corner_2 ? height - max_width_radius[ST_CORNER_BOTTOMLEFT]
+                         : height - border_width;
+      cogl_rectangle (x1, y1, x2, y2);
     }
 
   /* corners */
-  if (node->corner_texture != COGL_INVALID_HANDLE)
+  if (max_border_radius > 0)
     {
-      material = cogl_material_new ();
-      cogl_material_set_layer (material, 0, node->corner_texture);
-      cogl_material_set_color4ub (material,
-                                  paint_opacity, paint_opacity, paint_opacity, paint_opacity);
+      for (corner_id = 0; corner_id < 4; corner_id++)
+        {
+          if (node->corner_material[corner_id] == COGL_INVALID_HANDLE)
+            continue;
 
-      cogl_set_source (material);
+          cogl_material_set_color4ub (node->corner_material[corner_id],
+                                      paint_opacity, paint_opacity,
+                                      paint_opacity, paint_opacity);
+          cogl_set_source (node->corner_material[corner_id]);
 
-      cogl_rectangle_with_texture_coords (0, 0, max_width_radius, max_width_radius, 0, 0, 0.5, 0.5);
-      cogl_rectangle_with_texture_coords (width - max_width_radius, 0, width, max_width_radius, 0.5, 0, 1, 0.5);
-      cogl_rectangle_with_texture_coords (width - max_width_radius, height - max_width_radius, width, height, 0.5, 0.5, 1, 1);
-      cogl_rectangle_with_texture_coords (0, height - max_width_radius, max_width_radius, height, 0, 0.5, 0.5, 1);
-
-      cogl_handle_unref (material);
+          switch (corner_id)
+            {
+              case ST_CORNER_TOPLEFT:
+                cogl_rectangle_with_texture_coords (0, 0,
+                                                    max_width_radius[ST_CORNER_TOPLEFT], max_width_radius[ST_CORNER_TOPLEFT],
+                                                    0, 0, 0.5, 0.5);
+                break;
+              case ST_CORNER_TOPRIGHT:
+                cogl_rectangle_with_texture_coords (width - max_width_radius[ST_CORNER_TOPRIGHT], 0,
+                                                    width, max_width_radius[ST_CORNER_TOPRIGHT],
+                                                    0.5, 0, 1, 0.5);
+                break;
+              case ST_CORNER_BOTTOMRIGHT:
+                cogl_rectangle_with_texture_coords (width - max_width_radius[ST_CORNER_BOTTOMRIGHT], height - max_width_radius[ST_CORNER_BOTTOMRIGHT],
+                                                    width, height,
+                                                    0.5, 0.5, 1, 1);
+                break;
+              case ST_CORNER_BOTTOMLEFT:
+                cogl_rectangle_with_texture_coords (0, height - max_width_radius[ST_CORNER_BOTTOMLEFT],
+                                                    max_width_radius[ST_CORNER_BOTTOMLEFT], height,
+                                                    0, 0.5, 0.5, 1);
+                break;
+            }
+        }
     }
 
   /* background color */
@@ -935,7 +794,80 @@ st_theme_node_paint_borders (StThemeNode           *node,
                             node->background_color.blue,
                             paint_opacity * node->background_color.alpha / 255);
 
-  if (border_radius > border_width)
+  /* We add padding to each corner, so that all corners end up as if they
+   * had a border-radius of max_border_radius, which allows us to treat
+   * corners as uniform further on.
+   */
+  for (corner_id = 0; corner_id < 4; corner_id++)
+    {
+      float verts[8];
+      int n_rects;
+
+      /* corner texture does not need padding */
+      if (max_border_radius == node->border_radius[corner_id])
+        continue;
+
+      n_rects = node->border_radius[corner_id] == 0 ? 1 : 2;
+
+      switch (corner_id)
+        {
+          case ST_CORNER_TOPLEFT:
+            verts[0] = border_width;
+            verts[1] = max_width_radius[ST_CORNER_TOPLEFT];
+            verts[2] = max_border_radius;
+            verts[3] = max_border_radius;
+            if (n_rects == 2)
+              {
+                verts[4] = max_width_radius[ST_CORNER_TOPLEFT];
+                verts[5] = border_width;
+                verts[6] = max_border_radius;
+                verts[7] = max_width_radius[ST_CORNER_TOPLEFT];
+              }
+            break;
+          case ST_CORNER_TOPRIGHT:
+            verts[0] = width - max_border_radius;
+            verts[1] = max_width_radius[ST_CORNER_TOPRIGHT];
+            verts[2] = width - border_width;
+            verts[3] = max_border_radius;
+            if (n_rects == 2)
+              {
+                verts[4] = width - max_border_radius;
+                verts[5] = border_width;
+                verts[6] = width - max_width_radius[ST_CORNER_TOPRIGHT];
+                verts[7] = max_width_radius[ST_CORNER_TOPRIGHT];
+              }
+            break;
+          case ST_CORNER_BOTTOMRIGHT:
+            verts[0] = width - max_border_radius;
+            verts[1] = height - max_border_radius;
+            verts[2] = width - border_width;
+            verts[3] = height - max_width_radius[ST_CORNER_BOTTOMRIGHT];
+            if (n_rects == 2)
+              {
+                verts[4] = width - max_border_radius;
+                verts[5] = height - max_width_radius[ST_CORNER_BOTTOMRIGHT];
+                verts[6] = width - max_width_radius[ST_CORNER_BOTTOMRIGHT];
+                verts[7] = height - border_width;
+              }
+            break;
+          case ST_CORNER_BOTTOMLEFT:
+            verts[0] = border_width;
+            verts[1] = height - max_border_radius;
+            verts[2] = max_border_radius;
+            verts[3] = height - max_width_radius[ST_CORNER_BOTTOMLEFT];
+            if (n_rects == 2)
+              {
+                verts[4] = max_width_radius[ST_CORNER_BOTTOMLEFT];
+                verts[5] = height - max_width_radius[ST_CORNER_BOTTOMLEFT];
+                verts[6] = max_border_radius;
+                verts[7] = height - border_width;
+              }
+            break;
+        }
+      cogl_rectangles (verts, n_rects);
+    }
+
+  if (max_border_radius > border_width)
     {
       /* Once we've drawn the borders and corners, if the corners are bigger
        * the the border width, the remaining area is shaped like
@@ -948,14 +880,14 @@ st_theme_node_paint_borders (StThemeNode           *node,
        * We draw it in 3 pieces - first the top and bottom, then the main
        * rectangle
        */
-      cogl_rectangle (border_radius, border_width,
-                      width - border_radius, border_radius);
-      cogl_rectangle (border_radius, height - border_radius,
-                      width - border_radius, height - border_width);
+      cogl_rectangle (max_border_radius, border_width,
+                      width - max_border_radius, max_border_radius);
+      cogl_rectangle (max_border_radius, height - max_border_radius,
+                      width - max_border_radius, height - border_width);
     }
 
-  cogl_rectangle (border_width, max_width_radius,
-                  width - border_width, height - max_width_radius);
+  cogl_rectangle (border_width, MAX(border_width, max_border_radius),
+                  width - border_width, height - MAX(border_width, max_border_radius));
 }
 
 static void
@@ -992,8 +924,7 @@ st_theme_node_paint_sliced_border_image (StThemeNode           *node,
   if (ey < 0)
     ey = border_bottom;          /* FIXME ? */
 
-  material = cogl_material_new ();
-  cogl_material_set_layer (material, 0, node->border_texture);
+  material = node->border_material;
   cogl_material_set_color4ub (material,
                               paint_opacity, paint_opacity, paint_opacity, paint_opacity);
 
@@ -1050,8 +981,6 @@ st_theme_node_paint_sliced_border_image (StThemeNode           *node,
 
     cogl_rectangles_with_texture_coords (rectangles, 9);
   }
-
-  cogl_handle_unref (material);
 }
 
 static void
@@ -1153,16 +1082,16 @@ st_theme_node_paint (StThemeNode           *node,
    */
 
   if (node->border_shadow_material)
-    paint_shadow_with_opacity (node->border_shadow_material,
-                               node->shadow,
-                               &allocation,
-                               paint_opacity);
+    _st_paint_shadow_with_opacity (node->shadow,
+                                   node->border_shadow_material,
+                                   &allocation,
+                                   paint_opacity);
 
-  if (node->border_texture != COGL_INVALID_HANDLE)
+  if (node->border_material != COGL_INVALID_HANDLE)
     {
       /* Gradients and border images are mutually exclusive at this time */
       if (node->background_gradient_type != ST_GRADIENT_NONE)
-        paint_texture_with_opacity (node->border_texture, &allocation, paint_opacity);
+        paint_material_with_opacity (node->border_material, &allocation, paint_opacity);
       else
         st_theme_node_paint_sliced_border_image (node, &allocation, paint_opacity);
     }
@@ -1190,11 +1119,55 @@ st_theme_node_paint (StThemeNode           *node,
        * like boder and background color into account).
        */
       if (node->background_shadow_material != COGL_INVALID_HANDLE)
-        paint_shadow_with_opacity (node->background_shadow_material,
-                                   node->shadow,
-                                   &background_box,
-                                   paint_opacity);
+        _st_paint_shadow_with_opacity (node->shadow,
+                                       node->background_shadow_material,
+                                       &background_box,
+                                       paint_opacity);
 
-      paint_texture_with_opacity (node->background_texture, &background_box, paint_opacity);
+      paint_material_with_opacity (node->background_material, &background_box, paint_opacity);
     }
 }
+
+/**
+ * st_theme_node_copy_cached_paint_state:
+ * @node: a #StThemeNode
+ * @other: a different #StThemeNode
+ *
+ * Copy cached painting state from @other to @node. This function can be used to
+ * optimize redrawing cached background images when the style on an element changess
+ * in a way that doesn't affect background drawing. This function must only be called
+ * if st_theme_node_paint_equal (node, other) returns %TRUE.
+ */
+void
+st_theme_node_copy_cached_paint_state (StThemeNode *node,
+                                       StThemeNode *other)
+{
+  int corner_id;
+
+  g_return_if_fail (ST_IS_THEME_NODE (node));
+  g_return_if_fail (ST_IS_THEME_NODE (other));
+
+  /* Check omitted for speed: */
+  /* g_return_if_fail (st_theme_node_paint_equal (node, other)); */
+
+  _st_theme_node_free_drawing_state (node);
+
+  node->alloc_width = other->alloc_width;
+  node->alloc_height = other->alloc_height;
+
+  if (other->background_shadow_material)
+    node->background_shadow_material = cogl_handle_ref (other->background_shadow_material);
+  if (other->border_shadow_material)
+    node->border_shadow_material = cogl_handle_ref (other->border_shadow_material);
+  if (other->background_texture)
+    node->background_texture = cogl_handle_ref (other->background_texture);
+  if (other->background_material)
+    node->background_material = cogl_handle_ref (other->background_material);
+  if (other->border_texture)
+    node->border_texture = cogl_handle_ref (other->border_texture);
+  if (other->border_material)
+    node->border_material = cogl_handle_ref (other->border_material);
+  for (corner_id = 0; corner_id < 4; corner_id++)
+    if (other->corner_material[corner_id])
+      node->corner_material[corner_id] = cogl_handle_ref (other->corner_material[corner_id]);
+}
